static char RcsId[] = "$Header: /home/surya/will/src/extract/RCS/main.c,v 1.4 1994/02/10 16:55:00 will Exp $";
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>

/* This program implements `replace' and `extract' commands.  If the
 * last characters of argv[0] match the #defined string EXTRACT [REPLACE],
 * then the command is assumed to be extract [replace]; otherwise, it's
 * an error.
 */
#ifndef EXTRACT
#define EXTRACT "extract"
#endif
#ifndef REPLACE
#define REPLACE "replace"
#endif

char *prog;
char *file;
extern errno;

/***************************************************************************/
main(argc, argv)
int argc;
char **argv;
{
#define	ESIZE 200
#define BUFSIZE 1024*16

    int i, k, fd;
    long first, last, nstep, size;
    FILE *fp;
    char *s, *buf, *malloc(), *realloc();
    char mode[4];
    int is_plain;		/* != 0 if file is plain */
    int is_extract;		/* != 0 if function is extraction */
    int to_eof=0;		/* != 0 if range has $-sign */
    int count_from=1;	/* how to count: label first byte `count_from' */

    struct stat filestat;
    void extract(), replace();

    prog = ((s=strrchr(argv[0], '/')) && *(s+1)) ? s+1 : argv[0];

    /* Determine if this is extraction or replacement */
    is_extract = strcmp(prog+strlen(prog)-strlen(EXTRACT), EXTRACT) == 0;
    if (is_extract) {
	(void) strcpy(mode, "r");
    } else if (strcmp(prog+strlen(prog)-strlen(REPLACE), REPLACE) == 0) {
	(void) strcpy(mode, "r+");
    } else {
	(void) fprintf(stderr, "The command name `%s' isn't `%s' or `%s', so I\n",
				prog, EXTRACT, REPLACE);
	(void) fprintf(stderr, "\tdon't know whether to extract or replace!\n");
	(void) fprintf(stderr,
			"\t(See the source code, macros EXTRACT and REPLACE)\n");
	exit(1);
    }

    /* Check for the only option: -nnn */
    if (argc > 1 && *argv[1] == '-' && strlen(argv[1]) > 1) {
	count_from = atoi(argv[1]+1);
	if (count_from < 0) {
	    (void) fprintf(stderr, "%s: the count_from position %d may not\
be negative.\n", prog, count_from);
	    exit(1);
	}
	argc--;
	argv++;
    }

    if (argc < 3) {
	(void) fprintf(stderr, "Usage: %s [-n] range... file\n", prog);
	exit(1);
    }

    /* open file to be replaced/extracted */
    file = argv[--argc];
    argv[argc] = (char *) 0;
    if (strcmp(file, "-") == 0 && is_extract)
	fp = stdin;
    else if ((fp = fopen(file, mode)) == NULL) {
	(void) fprintf(stderr, "%s: couldn't open file %s: ", prog, file);
	perror("");
	exit(1);
    }
    fd = fileno(fp);

    /* Get filesize */
    if (fstat(fd, &filestat) < 0) {
	(void) fprintf(stderr, "%s: couldn't fstat file %s: ", prog, file);
	perror("");
	exit(1);
    }
    size = filestat.st_size;
    is_plain = ( (int) filestat.st_mode & 0170000 ) == S_IFREG;

    if ((buf = malloc(BUFSIZE)) == NULL) {
	(void) fprintf(stderr, "%s: couldn't malloc i/o buffer: ", prog);
	perror("");
	exit(1);
    }

    /* For each range: do extractions/replacements */
    for (i=1; i < argc; i++) {
	to_eof=0;
	k = getrange(argv[i], &first, &last, &nstep);
	switch (k) {
	case -1:
	    (void) fprintf(stderr, "%s: internal error\n", prog);
	    exit(1);
	    break;
	case 1:
	    (void) fprintf(stderr, "%s: invalid range `%s'\n", prog, argv[i]);
	    exit(1);
	    break;
	case 2:
	    (void) fprintf(stderr, "%s: in range `%s', the last position \
%d exceeds the first position %d.\n", prog, argv[i], last, first);
	    exit(1);
	    break;
	case 3:
	    (void) fprintf(stderr, "%s: in range `%s', the length %d \
must be positive.\n", prog, argv[i], last+1-first);
	    exit(1);
	    break;
	case 4:
	    (void) fprintf(stderr, "%s: in range `%s', the step size %d \
must be positive.\n", prog, argv[i], nstep);
	    exit(1);
	    break;
	default:
	    /* success */
	    break;
	}
	if (last < 0) {
	    to_eof = 1;
	    last = (is_plain) ? size+count_from-1 : LONG_MAX;
	}

	if (first < count_from) {
	    (void) fprintf(stderr, "%s: invalid range `%s': the first \
position given, %d, is < %d, the beginning of the file.\n",
		prog, argv[i], first, count_from);
	    exit(1);
	} else if (is_plain && first+1-count_from > size) {
	    (void) fprintf(stderr, "%s: invalid range `%s': the first \
position %d exceeds the file size %d.\n", prog, argv[i], first, size);
	    exit(1);
	}

	if (is_plain && last+1-count_from > size) {
	    (void) fprintf(stderr, "%s: in range `%s', the last \
position %d exceeds the file size %d.\n", prog, argv[i], last, size);
	    exit(1);
	}

	first += 1 - count_from;
	last += 1 - count_from;
	if (is_extract)
	    extract(fp, is_plain, buf, BUFSIZE, argv[i],
				first, last, nstep, size, to_eof);
	else
	    replace(fp, buf, BUFSIZE, argv[i], first, last, nstep, size);
    }
}
