static char RcsId[] = "$Header: extract.c,v 1.1 93/10/29 22:54:43 will Exp $";
#include <stdio.h>

extern char *prog;


/****************************************************************************/
void extract(fp, is_plain, buf, bufsize, range,
				first, last, nstep, size, to_eof)
/****************************************************************************/

	/* Extracts a range of data from stream and writes it on stdout */

FILE *fp;	/* File for extraction */
int is_plain;	/* specifies if file is plain */
char *buf;	/* Copy buffer */
int bufsize;	/* Copy buffer size */
char *range;	/* Range string entered by user; used for error messages */
int first;	/* first byte to copy (first byte in file is #1) */
int last;	/* last byte in range to copy (not including repetitions) */
		/* If to_eof is set, then last is set to a huge number, */
		/* and is pretty irrelevant. */
int nstep;	/* Step size for repetitions of first-last */
int size;	/* File size (if plain file); used for stopping repetitions */
int to_eof;	/* != 0 if range has $-sign */

{
    int k, len, toread, nread;
    static last_read = 0;	/* To get to proper spot on non-lseek files */

    len = last + 1 - first;

    /* Following loop is executed once for each nstep */
    /* Note: first is incremented at bottom of the do while loop;
     * hence it always contains the beginning of the next range
     * to read when nstep > 0.
     */
    do {
	/* Seek to beginning of range on plain files; */
	/* read to beginning on all others */
	if (is_plain) {
	    if (first-1 >= size)
		return;
	    if (fseek(fp, (long) (first-1), 0) == -1) {
		(void) fprintf(stderr,
			"%s: failed to lseek to beginning of range `%s': ",
			prog, range);
		perror("");
		exit(1);
	    }
	} else {
	    /* read to beginning */
	    if (last_read >= first) {
		(void) fprintf(stderr, "%s: Ranges must be monotonically \
increasing for non-plain files (see range %s).\n", prog, range);
		exit(1);
	    }
	    for (nread=last_read + 1; nread < first; ) {
		toread = first-nread;
		toread = (toread > bufsize) ? bufsize : toread;
		if ((k = fread(buf, 1,  toread, fp)) > 0) {
		    nread += k;
		    last_read += k;
		} else if (feof(fp) && nstep > 0) {
		    clearerr(fp);
		    fflush(stdout);
		    return;
		} else {
		    (void) fprintf(stderr,
			"%s: couldn't skip enough data to reach range `%s'.\n",
					prog, range);
		    if (feof(fp))
			(void) fprintf(stderr, "EOF encountered\n");
		    else
			perror("");
		    exit(1);
		}
	    }
	}

	for (nread=0; nread < len; ) {
	    toread = len-nread;
	    toread = (toread > bufsize) ? bufsize : toread;

	    if ((k = fread(buf, 1,  toread, fp)) > 0) {
		nread += k;
		last_read += k;
	    } else if (feof(fp) && (to_eof || nstep > 0)) {
		/* reached EOF */
		clearerr(fp);
		fflush(stdout);
		return;
	    } else if (feof(fp)) {
		(void) fprintf(stderr,
		"%s: reached EOF while reading data for range request `%s': ",
				    prog, range);
		exit(1);
	    } else if (ferror(fp)) {
		(void) fprintf(stderr,
		    "%s: Error while reading data for range request `%s'",
		    prog, range);
		if (nread)
		    (void) fprintf(stderr, " -- only read %d chars.: ", nread);
		else
		    (void) fprintf(stderr, ": ");
		perror("");
		exit(1);
	    }
	    if (fwrite(buf, 1, k, stdout) != k) {
		(void) fprintf(stderr,
			"%s: couldn't write all data from range request `%s': ",
				prog, range);
		perror("");
		exit(1);
	    }
	}
	first += nstep;
	last += nstep;
    } while (nstep > 0);
}
